#include <stdlib.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/draggers/SoTranslate2Dragger.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoSpotLight.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/nodes/SoNode.h>


#include "hyperboloid.h"
#include "rook.h"
#include "pawn.h"
#include "queen.h"
#include "king.h"
#include "bishop.h"
#include "knight.h"
#include "chessboard.h"

#include "SocketMan.h"
#include "SocketClient.h"

const int CHESS_PORT = 17283;
SocketMan *sm = NULL;
SocketClient *sc = NULL;


void send_movesCB(void *, SoDragger* dragger)
{
  static SocketBuf sbuf;
  char buf[256];

  SoTranslate2Dragger *sel_drag;
  sel_drag = (SoTranslate2Dragger *) dragger;

  SbName drag_name = sel_drag->getName();
  const char *drag_name_str = drag_name.getString();

  SbVec3f vec_drag = sel_drag->translation.getValue();
  float xpos = vec_drag[0];
  float ypos = vec_drag[1];
  float zpos = vec_drag[2];

  sprintf(buf, "%s %.3f %.3f %.3f", drag_name_str, xpos, ypos, zpos); 
  sbuf.setBufFromString(buf);

  if (sm)
    {
      sm->rewindSockets();
      while (sm->nextSocket())
	{
	  sm->writeWithTimeout(&sbuf);
	}
    }
  else
    {
      sc->writeWithTimeout(&sbuf);
    }

  printf("\n%s", buf);
  fflush(stdout);
}

void
movePiece(char *pieceName, float xPos, float yPos, float zPos)
{

  SoTranslate2Dragger *moved_drag;
  moved_drag = (SoTranslate2Dragger *) SoNode::getByName(SbName(pieceName));
  moved_drag->translation.setValue(xPos, yPos, zPos);
}  


void
networkCB(void *, SoSensor *sensor)
{
  static char pieceName[8];
  static float xPos, yPos, zPos;

  if (sm)
    {
      sm->acceptConnectionWithPoll();
      // Iterate over all clients, reading data if necessary
      sm->rewindSockets();
      while (sm->nextSocket())
	{
	  while (1)
	    {
	      SocketBuf *sbuf = sm->readWithPollAndTimeout();
	      if (sbuf == NULL)
		break;

	      sscanf(sbuf->buf, "%s %f %f %f",
		     pieceName, &xPos, &yPos, &zPos);
	      movePiece(pieceName, xPos, yPos, zPos);
	      delete sbuf;
	    }
	}
    }
  else
    {
      while (1)
	{
	  SocketBuf *sbuf = sc->readWithPollAndTimeout();
	  if (sbuf == NULL)
	    break;

          sscanf(sbuf->buf, "%s %f %f %f",
	         pieceName, &xPos, &yPos, &zPos);
          movePiece(pieceName, xPos, yPos, zPos);	  
	  delete sbuf;
	}
    }

  sensor->schedule();
}



void
usage()
{
  printf("Usage: Chess [-texture] [-connect <server name>]\n");
}

int main(int argc, char **argv)	{

  char *serverName = NULL;

  int flag_texture = 0;   // flag to indicate whether texture option is on

// Initialize Inventor

  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) exit(1);

// Parsing
  int c1 = 1;
  while (c1 < argc)
    {
      if (strcmp(argv[c1], "-texture") == 0)
	flag_texture = 1;
      else if (!strcmp(argv[c1], "-connect"))
	{
	  if ((c1+1) < argc)
	    {
	      c1++;
	      serverName = argv[c1];
	    }
	  else
	    {
	      usage();
	      exit(1);
	    }
	}
      else
	{
	  usage();
	  exit(1);
	}
      c1++;
    }

  // Set up the networking
  if (serverName)
    {
      // We are a client to a server
      sc = new SocketClient(CHESS_PORT);
      while (sc->connectToServer(serverName) !=
	     SocketClient::MSG_NO_ERROR)
	{
	  sleep(1);
	}
    }
  else
    {
      // We are a server to clients (we expect only one, though)
      sm = new SocketMan(CHESS_PORT);
    }

  SoIdleSensor *networkSensor = 
    new SoIdleSensor(networkCB, NULL);
  networkSensor->schedule();

  SoSeparator *root = new SoSeparator;
  root->ref();

// Add a Spotlight for general illumination
  SoSpotLight *light = new SoSpotLight;
  light->location.setValue(140,140,280);
  light->direction.setValue(0,0,-1);
  light->cutOffAngle = 1.0;
  light->intensity = 1.0;


// Chess Board Separator
  int board_block_size = 40;
  chessboard board(board_block_size, flag_texture);
  root->addChild(board.getGeometry());

// Transform from the chessboard to the surface of it, in place and 
// ready for the pieces.

  SoTransform *trans_to_surface = new SoTransform;
  trans_to_surface->translation.setValue(0, 0, board_block_size/4);
  root->addChild(trans_to_surface);

// White pieces
  // Separator for whites will have texture (if specified)  and then 
  // draggers for each of the pieces

  SoSeparator *whites = new SoSeparator;

  //Texture for white pieces
  SoTexture2 *texture_white = new SoTexture2;
  texture_white->filename.setValue("Yew.rgb");
  if (flag_texture)
    whites->addChild(texture_white);


  // White Rook 1
  rook wr1(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_rook_1 = new SoTranslate2Dragger;
  drag_white_rook_1->setPart("translator", wr1.getGeometry());
  drag_white_rook_1->setPart("translatorActive", wr1.getGeometry());
  drag_white_rook_1->translation.setValue(0,0,0);
  drag_white_rook_1->setName(SbName("wr1"));
  drag_white_rook_1->addMotionCallback(send_movesCB, NULL);

  // White Knight 1
  knight wn1(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_knight_1 = new SoTranslate2Dragger;
  drag_white_knight_1->setPart("translator", wn1.getGeometry());
  drag_white_knight_1->setPart("translatorActive", wn1.getGeometry());
  drag_white_knight_1->translation.setValue(40,0,0);
  drag_white_knight_1->setName(SbName("wn1"));
  drag_white_knight_1->addMotionCallback(send_movesCB, NULL);

  // White Bishop 1
  bishop wb1(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_bishop_1 = new SoTranslate2Dragger;
  drag_white_bishop_1->setPart("translator", wb1.getGeometry());
  drag_white_bishop_1->setPart("translatorActive", wb1.getGeometry());
  drag_white_bishop_1->translation.setValue(80,0,0);
  drag_white_bishop_1->setName(SbName("wb1"));
  drag_white_bishop_1->addMotionCallback(send_movesCB, NULL);

  // White Queen
  queen wq(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_queen = new SoTranslate2Dragger;
  drag_white_queen->setPart("translator", wq.getGeometry());
  drag_white_queen->setPart("translatorActive", wq.getGeometry());
  drag_white_queen->translation.setValue(120,0,0);
  drag_white_queen->setName(SbName("wq1"));
  drag_white_queen->addMotionCallback(send_movesCB, NULL);

  // White King
  king wk(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_king = new SoTranslate2Dragger;
  drag_white_king->setPart("translator", wk.getGeometry());
  drag_white_king->setPart("translatorActive", wk.getGeometry());
  drag_white_king->translation.setValue(160,0,0);
  drag_white_king->setName(SbName("wk1"));
  drag_white_king->addMotionCallback(send_movesCB, NULL);
  
  // White Bishop 2
  bishop wb2(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_bishop_2 = new SoTranslate2Dragger;
  drag_white_bishop_2->setPart("translator", wb2.getGeometry());
  drag_white_bishop_2->setPart("translatorActive", wb2.getGeometry());
  drag_white_bishop_2->translation.setValue(200,0,0);
  drag_white_bishop_2->setName(SbName("wb2"));
  drag_white_bishop_2->addMotionCallback(send_movesCB, NULL);

  // White Knight 2
  knight wn2(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_knight_2 = new SoTranslate2Dragger;
  drag_white_knight_2->setPart("translator", wn2.getGeometry());
  drag_white_knight_2->setPart("translatorActive", wn2.getGeometry());
  drag_white_knight_2->translation.setValue(240,0,0);
  drag_white_knight_2->setName(SbName("wn2"));
  drag_white_knight_2->addMotionCallback(send_movesCB, NULL);

  // White Rook 2
  rook wr2(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_rook_2 = new SoTranslate2Dragger;
  drag_white_rook_2->setPart("translator", wr2.getGeometry());
  drag_white_rook_2->setPart("translatorActive", wr2.getGeometry());
  drag_white_rook_2->translation.setValue(280,0,0);
  drag_white_rook_2->setName(SbName("wr2"));
  drag_white_rook_2->addMotionCallback(send_movesCB, NULL);

  // White Pawn 1
  pawn wp1(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_1 = new SoTranslate2Dragger;
  drag_white_pawn_1->setPart("translator", wp1.getGeometry());
  drag_white_pawn_1->setPart("translatorActive", wp1.getGeometry());
  drag_white_pawn_1->translation.setValue(0,40,0);
  drag_white_pawn_1->setName(SbName("wp1"));
  drag_white_pawn_1->addMotionCallback(send_movesCB, NULL);

  // White Pawn 2
  pawn wp2(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_2 = new SoTranslate2Dragger;
  drag_white_pawn_2->setPart("translator", wp2.getGeometry());
  drag_white_pawn_2->setPart("translatorActive", wp2.getGeometry());
  drag_white_pawn_2->translation.setValue(40,40,0);
  drag_white_pawn_2->setName(SbName("wp2"));
  drag_white_pawn_2->addMotionCallback(send_movesCB, NULL);

  // White Pawn 3
  pawn wp3(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_3 = new SoTranslate2Dragger;
  drag_white_pawn_3->setPart("translator", wp3.getGeometry());
  drag_white_pawn_3->setPart("translatorActive", wp3.getGeometry());
  drag_white_pawn_3->translation.setValue(80,40,0);
  drag_white_pawn_3->setName(SbName("wp3"));
  drag_white_pawn_3->addMotionCallback(send_movesCB, NULL);

  // White Pawn 4
  pawn wp4(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_4 = new SoTranslate2Dragger;
  drag_white_pawn_4->setPart("translator", wp4.getGeometry());
  drag_white_pawn_4->setPart("translatorActive", wp4.getGeometry());
  drag_white_pawn_4->translation.setValue(120,40,0);
  drag_white_pawn_4->setName(SbName("wp4"));
  drag_white_pawn_4->addMotionCallback(send_movesCB, NULL);

  // White Pawn 5
  pawn wp5(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_5 = new SoTranslate2Dragger;
  drag_white_pawn_5->setPart("translator", wp5.getGeometry());
  drag_white_pawn_5->setPart("translatorActive", wp5.getGeometry());
  drag_white_pawn_5->translation.setValue(160,40,0);
  drag_white_pawn_5->setName(SbName("wp5"));
  drag_white_pawn_5->addMotionCallback(send_movesCB, NULL);

  // White Pawn 6
  pawn wp6(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_6 = new SoTranslate2Dragger;
  drag_white_pawn_6->setPart("translator", wp6.getGeometry());
  drag_white_pawn_6->setPart("translatorActive", wp6.getGeometry());
  drag_white_pawn_6->translation.setValue(200,40,0);
  drag_white_pawn_6->setName(SbName("wp6"));
  drag_white_pawn_6->addMotionCallback(send_movesCB, NULL);

  // White Pawn 7
  pawn wp7(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_7 = new SoTranslate2Dragger;
  drag_white_pawn_7->setPart("translator", wp7.getGeometry());
  drag_white_pawn_7->setPart("translatorActive", wp7.getGeometry());
  drag_white_pawn_7->translation.setValue(240,40,0);
  drag_white_pawn_7->setName(SbName("wp7"));
  drag_white_pawn_7->addMotionCallback(send_movesCB, NULL);

  // White Pawn 8
  pawn wp8(1.0,1.0,1.0);
  
  SoTranslate2Dragger *drag_white_pawn_8 = new SoTranslate2Dragger;
  drag_white_pawn_8->setPart("translator", wp8.getGeometry());
  drag_white_pawn_8->setPart("translatorActive", wp8.getGeometry());
  drag_white_pawn_8->translation.setValue(280,40,0);
  drag_white_pawn_8->setName(SbName("wp8"));
  drag_white_pawn_8->addMotionCallback(send_movesCB, NULL);


  // Add Draggers to the Whites Root node
  whites->addChild(drag_white_rook_1);
  whites->addChild(drag_white_knight_1);
  whites->addChild(drag_white_bishop_1);
  whites->addChild(drag_white_queen);
  whites->addChild(drag_white_king);
  whites->addChild(drag_white_bishop_2);
  whites->addChild(drag_white_knight_2);
  whites->addChild(drag_white_rook_2);
  whites->addChild(drag_white_pawn_1);
  whites->addChild(drag_white_pawn_2);
  whites->addChild(drag_white_pawn_3);
  whites->addChild(drag_white_pawn_4);
  whites->addChild(drag_white_pawn_5);
  whites->addChild(drag_white_pawn_6);
  whites->addChild(drag_white_pawn_7);
  whites->addChild(drag_white_pawn_8);

// Black pieces
  // Separator for blacks will have texture (if specified) and then 
  // draggers for each of the pieces
  SoSeparator *blacks = new SoSeparator;

  //Texture for black pieces
  SoTexture2 *texture_black = new SoTexture2;
  texture_black->filename.setValue("Darkwd.rgb");
  // Set default color of pieces to be black
  float r = 0.1; float g = 0.1; float b = 0.1;

  if (flag_texture)
    {
      blacks->addChild(texture_black);
      r = 1.0; g = 1.0; b = 1.0;   // Modulate texture color with white
    }
  // Black Rook 1
  rook br1(r,g,b);
  
  SoTranslate2Dragger *drag_black_rook_1 = new SoTranslate2Dragger;
  drag_black_rook_1->setPart("translator", br1.getGeometry());
  drag_black_rook_1->setPart("translatorActive", br1.getGeometry());
  drag_black_rook_1->translation.setValue(0,280,0);
  drag_black_rook_1->setName(SbName("br1"));
  drag_black_rook_1->addMotionCallback(send_movesCB, NULL);

  // Black Knight 1
  knight bn1(r,g,b);
  
  SoTranslate2Dragger *drag_black_knight_1 = new SoTranslate2Dragger;
  drag_black_knight_1->setPart("translator", bn1.getGeometry());
  drag_black_knight_1->setPart("translatorActive", bn1.getGeometry());
  drag_black_knight_1->translation.setValue(40,280,0);
  drag_black_knight_1->setName(SbName("bn1"));
  drag_black_knight_1->addMotionCallback(send_movesCB, NULL);

  // Black Bishop 1
  bishop bb1(r,g,b);
  
  SoTranslate2Dragger *drag_black_bishop_1 = new SoTranslate2Dragger;
  drag_black_bishop_1->setPart("translator", bb1.getGeometry());
  drag_black_bishop_1->setPart("translatorActive", bb1.getGeometry());
  drag_black_bishop_1->translation.setValue(80,280,0);
  drag_black_bishop_1->setName(SbName("bb1"));
  drag_black_bishop_1->addMotionCallback(send_movesCB, NULL);

  // Black Queen

  queen bq(r,g,b);
  
  SoTranslate2Dragger *drag_black_queen = new SoTranslate2Dragger;
  drag_black_queen->setPart("translator", bq.getGeometry());
  drag_black_queen->setPart("translatorActive", bq.getGeometry());
  drag_black_queen->translation.setValue(120,280,0);
  drag_black_queen->setName(SbName("bq1"));
  drag_black_queen->addMotionCallback(send_movesCB, NULL);

  // Black King
  king bk(r,g,b);
  
  SoTranslate2Dragger *drag_black_king = new SoTranslate2Dragger;
  drag_black_king->setPart("translator", bk.getGeometry());
  drag_black_king->setPart("translatorActive", bk.getGeometry());
  drag_black_king->translation.setValue(160,280,0);
  drag_black_king->setName(SbName("bk1"));
  drag_black_king->addMotionCallback(send_movesCB, NULL);

  // Black Bishop 2
  bishop bb2(r,g,b);
  
  SoTranslate2Dragger *drag_black_bishop_2 = new SoTranslate2Dragger;
  drag_black_bishop_2->setPart("translator", bb2.getGeometry());
  drag_black_bishop_2->setPart("translatorActive", bb2.getGeometry());
  drag_black_bishop_2->translation.setValue(200,280,0);
  drag_black_bishop_2->setName(SbName("bb2"));
  drag_black_bishop_2->addMotionCallback(send_movesCB, NULL);

  // Black Knight 2
  knight bn2(r,g,b);
  
  SoTranslate2Dragger *drag_black_knight_2 = new SoTranslate2Dragger;
  drag_black_knight_2->setPart("translator", bn2.getGeometry());
  drag_black_knight_2->setPart("translatorActive", bn2.getGeometry());
  drag_black_knight_2->translation.setValue(240,280,0);
  drag_black_knight_2->setName(SbName("bn2"));
  drag_black_knight_2->addMotionCallback(send_movesCB, NULL);

  // Black Rook 2
  rook br2(r,g,b);
  
  SoTranslate2Dragger *drag_black_rook_2 = new SoTranslate2Dragger;
  drag_black_rook_2->setPart("translator", br2.getGeometry());
  drag_black_rook_2->setPart("translatorActive", br2.getGeometry());
  drag_black_rook_2->translation.setValue(280,280,0);
  drag_black_rook_2->setName(SbName("br2"));
  drag_black_rook_2->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 1
  pawn bp1(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_1 = new SoTranslate2Dragger;
  drag_black_pawn_1->setPart("translator", bp1.getGeometry());
  drag_black_pawn_1->setPart("translatorActive", bp1.getGeometry());
  drag_black_pawn_1->translation.setValue(0,240,0);
  drag_black_pawn_1->setName(SbName("bp1"));
  drag_black_pawn_1->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 2
  pawn bp2(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_2 = new SoTranslate2Dragger;
  drag_black_pawn_2->setPart("translator", bp2.getGeometry());
  drag_black_pawn_2->setPart("translatorActive", bp2.getGeometry());
  drag_black_pawn_2->translation.setValue(40,240,0);
  drag_black_pawn_2->setName(SbName("bp2"));
  drag_black_pawn_2->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 3
  pawn bp3(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_3 = new SoTranslate2Dragger;
  drag_black_pawn_3->setPart("translator", bp3.getGeometry());
  drag_black_pawn_3->setPart("translatorActive", bp3.getGeometry());
  drag_black_pawn_3->translation.setValue(80,240,0);
  drag_black_pawn_3->setName(SbName("bp3"));
  drag_black_pawn_3->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 4
  pawn bp4(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_4 = new SoTranslate2Dragger;
  drag_black_pawn_4->setPart("translator", bp4.getGeometry());
  drag_black_pawn_4->setPart("translatorActive", bp4.getGeometry());
  drag_black_pawn_4->translation.setValue(120,240,0);
  drag_black_pawn_4->setName(SbName("bp4"));
  drag_black_pawn_4->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 5
  pawn bp5(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_5 = new SoTranslate2Dragger;
  drag_black_pawn_5->setPart("translator", bp5.getGeometry());
  drag_black_pawn_5->setPart("translatorActive", bp5.getGeometry());
  drag_black_pawn_5->translation.setValue(160,240,0);
  drag_black_pawn_5->setName(SbName("bp5"));
  drag_black_pawn_5->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 6
  pawn bp6(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_6 = new SoTranslate2Dragger;
  drag_black_pawn_6->setPart("translator", bp6.getGeometry());
  drag_black_pawn_6->setPart("translatorActive", bp6.getGeometry());
  drag_black_pawn_6->translation.setValue(200,240,0);
  drag_black_pawn_6->setName(SbName("bp6"));
  drag_black_pawn_6->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 7
  pawn bp7(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_7 = new SoTranslate2Dragger;
  drag_black_pawn_7->setPart("translator", bp7.getGeometry());
  drag_black_pawn_7->setPart("translatorActive", bp7.getGeometry());
  drag_black_pawn_7->translation.setValue(240,240,0);
  drag_black_pawn_7->setName(SbName("bp7"));
  drag_black_pawn_7->addMotionCallback(send_movesCB, NULL);

  // Black Pawn 8
  pawn bp8(r,g,b);
  
  SoTranslate2Dragger *drag_black_pawn_8 = new SoTranslate2Dragger;
  drag_black_pawn_8->setPart("translator", bp8.getGeometry());
  drag_black_pawn_8->setPart("translatorActive", bp8.getGeometry());
  drag_black_pawn_8->translation.setValue(280,240,0);
  drag_black_pawn_8->setName(SbName("bp8"));
  drag_black_pawn_8->addMotionCallback(send_movesCB, NULL);


  // Add Draggers to the Blacks Root node
  blacks->addChild(drag_black_rook_1);
  blacks->addChild(drag_black_knight_1);
  blacks->addChild(drag_black_bishop_1);
  blacks->addChild(drag_black_queen);
  blacks->addChild(drag_black_king);
  blacks->addChild(drag_black_bishop_2);
  blacks->addChild(drag_black_knight_2);
  blacks->addChild(drag_black_rook_2);
  blacks->addChild(drag_black_pawn_1);
  blacks->addChild(drag_black_pawn_2);
  blacks->addChild(drag_black_pawn_3);
  blacks->addChild(drag_black_pawn_4);
  blacks->addChild(drag_black_pawn_5);
  blacks->addChild(drag_black_pawn_6);
  blacks->addChild(drag_black_pawn_7);
  blacks->addChild(drag_black_pawn_8);

  //root->addChild(chess_board);
  //root->addChild(trans_to_board_surface);
  root->addChild(whites);
  root->addChild(blacks);

// Set up veiwer / Render area to see the scene graph

  SoXtExaminerViewer *viewer = new SoXtExaminerViewer(myWindow);
  viewer->setSceneGraph(root);
  viewer->setTitle("Chess");
  viewer->show();
//  if (!flag_texture)
    viewer->setBackgroundColor(SbColor(0.12,0.0,0.15));

  SoXt::show(myWindow);
  SoXt::mainLoop();

}







